/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.api;

/**
 * 
 * All Classes should have an own Logger
 * 
 * There are Implementations for applications
 * which are logging itself (per ex. Logging4BB)
 * or which are logging through remote Commands
 * like LogManager4BB
 * 
 * Log Levels supported:
 * 
 * DEBUG
 * LOG or INFO (same level)
 * WARN
 * ERROR
 * 
 * logs can be logged to a LogMonitor like Lilith
 * streamed through a Socket Connection
 * 
 * or logged into SQLite Database
 * (this local Database cann be transmitted to Server)
 * 
 * or as a livestream to a WebSocket Connection
 * 
 * 
 * @author ekkehard gentz, rosenheim, germany
 *
 */
public interface ILogger {
	
	/**
	 * lowest Loglevel: Debug - should only be used while testing / debugging
	 */
	public static final int LOG_LEVEL_DEBUG = 0;
	/**
	 * default Loglevel: INFO or LOG
	 */
	public static final int LOG_LEVEL_INFO = 1;
	/**
	 * if you know that your App is running well,
	 * then you can also use WARN as your defuelt level
	 * WARN log events are im,portant to know,
	 * but don't stop the execution of tasks
	 */
	public static final int LOG_LEVEL_WARN = 2;
	/**
	 * ERROR log level should be watched 
	 * if an ERROR occured your application doesn't work as expected
	 */
	public static final int LOG_LEVEL_ERROR = 3;
	/**
	 * TEXT (String) representation of the Loglevels
	 */
	public static String[] logLevelTexts = {
		"DEBUG",
		"INFO",
		"WARN",
		"ERROR"
	};
	
	//// KEYS used for remote logging
	// org.appbus.bb.app.logging4bb.model.logstoreid
	// 0x2c89d895219a0725L
	public static final Long LOGSTOREID_KEY = new Long(0x2c89d895219a0725L);
	// org.appbus.bb.app.logging4bb.model.appname.db
	// 0x3d27694c4fa6692bL
	public static final Long APPNAMEDB_KEY = new Long(0x3d27694c4fa6692bL);
	// org.appbus.bb.app.logging4bb.model.appname.socket
	// 0x53896ffd3e580da1L
	public static final Long APPNAMESOCKET_KEY = new Long(0x53896ffd3e580da1L);
	// org.appbus.bb.app.logging4bb.model.loggername
	// 0xaf5f7369ab990261L
	public static final Long LOGGERNAME_KEY = new Long(0xaf5f7369ab990261L);
	// org.appbus.bb.app.logging4bb.model.context
	// 0xdbf5fbf014906f83L
	public static final Long CONTEXT_KEY = new Long(0xdbf5fbf014906f83L);
	// org.appbus.bb.app.logging4bb.model.message
	// 0x215b751f0233889dL
	public static final Long MESSAGE_KEY = new Long(0x215b751f0233889dL);
	// org.appbus.bb.app.logging4bb.model.level
	// 0x82f1fab9c87827b8L
	public static final Long LEVEL_KEY = new Long(0x82f1fab9c87827b8L);
	// org.appbus.bb.app.logging4bb.model.eventtime
	// 0xaafdb00ca056b2e8L
	public static final Long EVENTTIME_KEY = new Long(0xaafdb00ca056b2e8L);
	// org.appbus.bb.app.logging4bb.model.exeption
	// 0x33dc2775521e1a97L
	public static final Long EXCEPTION_KEY = new Long(0x33dc2775521e1a97L);

	
	/**
	 * logs message with default LEVEL INFO
	 * @param logMessage
	 */
	public void log(String logMessage);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with an Exception
	 * 
	 * @param exception
	 * @param logMessage
	 */
	public void log(String logMessage, Throwable exception);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with context (like method name)
	 * 
	 * @param context
	 * @param logMessage
	 */
	public void log(String context, String logMessage);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with context (like method name)
	 * and Exception
	 * 
	 * @param context
	 * @param logMessage
	 * @param exception
	 */
	public void log(String context, String logMessage, Throwable exception);
	
	/**
	 * logs message with default LEVEL INFO
	 * @param logMessage
	 */
	public void info(String logMessage);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with an Exception
	 * 
	 * @param exception
	 * @param logMessage
	 */
	public void info(String logMessage, Throwable exception);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with context (like method name)
	 * 
	 * @param context
	 * @param logMessage
	 */
	public void info(String context, String logMessage);
	
	/**
	 * logs message with default LEVEL INFO
	 * together with context (like method name)
	 * and Exception
	 * 
	 * @param context
	 * @param logMessage
	 * @param exception
	 */
	public void info(String context, String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL WARN
	 * 
	 * @param logMessage
	 */
	public void warn(String logMessage);
	
	/**
	 * logs message with default LEVEL WARN
	 * together with an Exception
	 * 
	 * @param exception
	 * @param logMessage
	 */
	public void warn(String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL WARN
	 * together with context (like method name)
	 * 
	 * @param context
	 * @param logMessage
	 */
	public void warn(String context, String logMessage);
	
	/**
	 * logs message with default LEVEL WARN
	 * together with context (like method name)
	 * and Exception
	 * 
	 * @param context
	 * @param logMessage
	 * @param exception
	 */
	public void warn(String context, String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL ERROR
	 * 
	 * @param logMessage
	 */
	public void error(String logMessage);
	
	/**
	 * logs message with default LEVEL ERROR
	 * together with an Exception
	 * 
	 * @param exception
	 * @param logMessage
	 */
	public void error(String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL ERROR
	 * together with context (like method name)
	 * 
	 * @param context
	 * @param logMessage
	 */
	public void error(String context, String logMessage);
	
	/**
	 * logs message with default LEVEL ERROR
	 * together with context (like method name)
	 * and Exception
	 * 
	 * @param context
	 * @param logMessage
	 * @param exception
	 */
	public void error(String context, String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL DEBUG
	 * 
	 * @param logMessage
	 */
	public void debug(String logMessage);
	
	/**
	 * logs message with default LEVEL DEBUG
	 * together with an Exception
	 * 
	 * @param exception
	 * @param logMessage
	 */
	public void debug(String logMessage, Throwable exception);
	
	/**
	 * logs message with LEVEL DEBUG
	 * together with context (like method name)
	 * 
	 * @param context
	 * @param logMessage
	 */
	public void debug(String context, String logMessage);
	
	/**
	 * logs message with default LEVEL DEBUG
	 * together with context (like method name)
	 * and Exception
	 * 
	 * @param context
	 * @param logMessage
	 * @param exception
	 */
	public void debug(String context, String logMessage, Throwable exception);
	
	/**
	 * true if DEBUG is enabled
	 * should be used to test if complex String concat etc was done
	 * 
	 * @return
	 */
	public boolean isDebugEnabled();
}
