/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.db;

import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import net.rim.device.api.system.LED;
import net.rim.device.api.ui.Color;
import net.rim.device.api.util.StringProvider;

import org.appbus.bb.app.logging4bb.CallbackNotificationTask;
import org.appbus.bb.app.logging4bb.CommandFactory;
import org.appbus.bb.app.logging4bb.Logging4BBApp;
import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;
import org.appbus.bb.app.logging4bb.model.LogEventTask;
import org.appbus.bb.app.logging4bb.util.Utility;


public class Logging2DatabaseWorker {
	
	private static Vector tasks = new Vector(Logging4BBApp.DEFAULT_LOG_TASKS_BUFFER_SIZE);
	private static boolean isWorking = false;
	private static boolean isPaused = false;
	private static boolean haveMoreWorkToDo = false;
	private static WorkerThread theThread;
	private static boolean bufferTooSmallLogged = false;
	
	/**
	 * adds one task to the Vector
	 * synchronized on the Vector, so data is always consistent
	 * 
	 * @param task
	 */
	public static void addTask(Object task){
		if (isPaused) {
			return;
		}
		synchronized (tasks) {
			if (tasks.size() < Logging4BBApp.getLogTasksBufferSize()) {
				tasks.addElement(task);
				bufferTooSmallLogged = false;
			} else {
				// ignore task
				// max reached log error (only one time)
				if (!bufferTooSmallLogged) {
					bufferTooSmallLogged = true;
					tasks.addElement(getErrorLogEventBufferTooSmall());
				}
			}
		}
	}
	
	/**
	 * adds a bunch of tasks to the Vector
	 * synchronized on the Vector, so data is always consistent
	 * 
	 * @param tasks
	 */
	public static void addTasks(Hashtable[] someTasks){
		if (isPaused) {
			return;
		}
		synchronized (tasks) {
			if ((tasks.size()+someTasks.length) <= Logging4BBApp.getLogTasksBufferSize()) {
				for (int i = 0; i < someTasks.length; i++) {
					tasks.addElement(new LogEventTask(someTasks[i]));
				}
				bufferTooSmallLogged = false;
			} else {
				int max = Logging4BBApp.getLogTasksBufferSize()-tasks.size();
				for (int i = 0; i <= max; i++) {
					tasks.addElement(new LogEventTask(someTasks[i]));
				}
				// ignore other tasks
				// max reached log error (only one time)
				if (!bufferTooSmallLogged) {
					bufferTooSmallLogged = true;
					tasks.addElement(getErrorLogEventBufferTooSmall());
				}
			}
		}
	}
	
	private static LogEventTask getErrorLogEventBufferTooSmall(){
		LogEventTask errorLogEvent = new LogEventTask(
				LoggerFactory.LOG_STORE_ID, 
				LoggerFactory.APP_NAME, 
				Logging2DatabaseWorker.class.getName());
		errorLogEvent.setContext("unexpected behaviour");
		errorLogEvent.setMessage("LogBuffer too small: "+Logging4BBApp.getLogTasksBufferSize());
		errorLogEvent.setLevel(ILogger.LOG_LEVEL_ERROR);
		return errorLogEvent;
	}
	
	/**
	 * tries to get the next task from the vector
	 * if Vector is empty, NoSuchElementException was thrown
	 * so the caller knows to go into sleep mode
	 * 
	 * @return
	 * @throws NoSuchElementException
	 */
	public static Object getNextTask() throws NoSuchElementException{
		if (tasks.isEmpty()) {
			throw new NoSuchElementException("Worker queue is empty");
		}
		synchronized (tasks) {
			Object task = (Object) tasks.firstElement();
			tasks.removeElementAt(0);
			return task;
		}
	}
	
	public static void startup(){
		isWorking = true;
		isPaused = false;
		theThread = new WorkerThread();
		theThread.start();
	}
	
	public static void ensureCapacity(int minimumCapacity){
		// tasks are created with minimum 500
		// this can be overwritten by startup parameters
		// or from LogManager
		synchronized (tasks) {
			tasks.ensureCapacity(minimumCapacity); 
		}
	}
	
	public static void shutdown(){
		if (theThread != null) {
			if (theThread.isAlive()) {
				theThread.interrupt();
			}
			theThread = null;
		}
	}

	/**
	 * pauses the Thread means:
	 * sleeps 3000 ms, then tests again if still paused and should not work
	 */
	public static void pauseWorker() {
		if (theThread != null && theThread.isAlive()) {
			isPaused = true;
			isWorking = false;
		}
	}

	/**
	 * restarts the worker:
	 * if Thread is alive: stop pause, start working
	 * otherwise: start a new Thread
	 * 
	 */
	public static void restartWorker(){
		if (!Logging4BBApp.logToDB()) {
			return;
		}
		if (theThread != null && theThread.isAlive()) {
			isWorking = true;
			isPaused = false;
		} else {
			startup();
		}
	}

	/**
	 * Thread tests if there are Tasks to do
	 * if no tasks are there sleep 1 sec and try again
	 * 
	 * if isWorking is set to false and no more work to be done
	 * the Thread terminates
	 * 
	 * @author ekke
	 *
	 */
	public static class WorkerThread extends Thread {
		public void run() {
			try {
				while ( isPaused || isWorking || haveMoreWorkToDo) {
					if (isPaused) {
						sleep(3000);
						if (isWorking) {
							isPaused = false;
						}
					} else {
						try {
							Object task = getNextTask();
							haveMoreWorkToDo = true;
							if (task instanceof CallbackNotificationTask) {
								callbackNotification((CallbackNotificationTask)task);
							} else if (task instanceof LogEventTask) {
								logEvent((LogEventTask)task);
							} else if (task instanceof String) {
								logStatus((String)task, ILogger.LOG_LEVEL_INFO);
							} else {
								logStatus("ERROR: wrong Task Instance", ILogger.LOG_LEVEL_ERROR);
							}
							
						} catch (NoSuchElementException e) {
							if (isWorking) {	
								// nothing to do - we sleep and try again
								sleep(1000);
							} else {
								haveMoreWorkToDo = false;
							}
						}
					}
				}
			} catch (InterruptedException e) {
				isWorking = false;
				isPaused = false;
			}
		}
		
		private static void logStatus(String message, int level){
			LogEventTask bean = new LogEventTask(
					LoggerFactory.LOG_STORE_ID, 
					LoggerFactory.APP_NAME, 
					Logging2DatabaseWorker.class.getName());
			bean.setContext("unexpected behaviour");
			bean.setMessage(message);
			bean.setEventTime(System.currentTimeMillis());
			bean.setLevel(level);
			try {
				Logging4BBDatabaseManager.insertLogEventIntoDatabase(bean);
			} catch (DatabaseOperationFailedException e) {
				// TODO notify user
			}
		}
		
		/**
		 * inserts the event into the DB
		 * if DEBUG_MODE is set, the LED blinks
		 * 
		 * @param bean
		 */
		private static void logEvent(LogEventTask bean){
			// blink LED if debug mode
			if (Logging4BBApp.isDebugMode()) {
				switch (bean.getLevel()) {
				case ILogger.LOG_LEVEL_DEBUG:
					Utility.useLED(Color.MAGENTA, LED.STATE_BLINKING);
					break;
				case ILogger.LOG_LEVEL_WARN:
					Utility.useLED(Color.ORANGE, LED.STATE_BLINKING);
					break;
				case ILogger.LOG_LEVEL_ERROR:
					Utility.useLED(Color.RED, LED.STATE_BLINKING);
					break;
				default:
					Utility.useLED(Color.GREEN, LED.STATE_BLINKING);
					break;
				}
			}
			// insert the event
			try {
				Logging4BBDatabaseManager.insertLogEventIntoDatabase(bean);
			} catch (DatabaseOperationFailedException e) {
				// TODO notify user ? send email ?
			}
			if (Logging4BBApp.isDebugMode()) {
				Utility.useLED(Color.BLACK, LED.STATE_OFF);
			}
		}
		
		
		private void callbackNotification(CallbackNotificationTask notificationtask){
			StringProvider content = new StringProvider(
					notificationtask.getTaskDescription().toString() + " --> work done !");
			CommandFactory.logging4bbCallbackCommandRemoteExecute(
					"logging4bbCallback",
					new String[]{notificationtask.getCommandCategory().toString()},
					new String[] {notificationtask.getContextCategory().toString()},
					null, // no context to give back
					content // message as content
			);
			
		}
		
	}

}
