/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.db;

import java.util.Hashtable;

import net.rim.device.api.database.Cursor;
import net.rim.device.api.database.DataTypeException;
import net.rim.device.api.database.Database;
import net.rim.device.api.database.DatabaseException;
import net.rim.device.api.database.DatabaseFactory;
import net.rim.device.api.database.DatabaseIOException;
import net.rim.device.api.database.DatabasePathException;
import net.rim.device.api.database.DatabaseSecurityOptions;
import net.rim.device.api.database.Row;
import net.rim.device.api.database.Statement;
import net.rim.device.api.io.MalformedURIException;
import net.rim.device.api.io.URI;
import net.rim.device.api.system.Characters;
import net.rim.device.api.system.ControlledAccessException;
import net.rim.device.api.util.LongHashtable;
import net.rim.device.api.util.LongVector;

import org.appbus.bb.app.logging4bb.CommandFactory;
import org.appbus.bb.app.logging4bb.Logging4BBApp;
import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;
import org.appbus.bb.app.logging4bb.model.LogEventTask;
import org.appbus.bb.app.logging4bb.util.Utility;

/**
 * Database operations can be called from foreground apps
 * via RemoteCommands or directly from inside this app
 * using some async Threads
 * 
 * Only ONE user can access the DB at a time - even if only read-only
 * (r/o - behaviour becomes better in BB OS 7)
 * 
 * We sync all calls using a DATABASE_SYNC_OBJECT
 * to avoid collisions
 * 
 * @author ekke
 *
 */
public class Logging4BBDatabaseManager {
	
	public static final ILogger logger = LoggerFactory.getLogger(Logging4BBDatabaseManager.class);
	
	// D A T A B A S E
	
	// database can be null if close or never opened
	private static Database logging4bbDatabase;
	private static boolean logging4bbDatabaseIsPaused = false;
	private static final Object DATABASE_SYNC_OBJECT = new Object(); 
	private static final boolean CRITICAL_ERROR = true;
	private static final boolean EXIT_THE_APP = true;
	private static final boolean WORK_ON = false;
	
    // Tables
    public static String CREATE_TABLE = "CREATE TABLE";
    public static String INSERT_INTO = "INSERT INTO";
    public static String VALUES = "VALUES";
    private static String UPDATE = "UPDATE";
    public static String DELETE_FROM = "DELETE FROM";
    public static String WHERE = "WHERE";
    public static String SET = "SET";
    public static String SELECT = "SELECT";
    public static String FROM = "FROM";
    public static String AND = "AND";
    public static String OR = "OR";
    public static String COUNT = "COUNT";
    public static String ORDER_BY = "ORDER BY";
    public static String ASC = "ASC";
    public static String DESC = "DESC";
    public static String LIMIT = "LIMIT";
    
    public static String COL_TYPE_TEXT = "TEXT";
    public static String COL_TYPE_NUMERIC = "NUMERIC";
    public static String COL_TYPE_PK = "INTEGER PRIMARY KEY";
    public static String COL_TYPE_DATETIME = "DATETIME";
    
    public static String TABLE_LOG_EVENT = "T_LOG_EVENT";
    public static String COL_PREFIX_LOG_EVENT = "LOG_";
    public static String COL_ID = "ID";
    public static String COL_DEVICE_ID = "DEVICE_ID";
    public static String COL_STORE_ID = "STORE_ID";
    public static String COL_APP_NAME = "APP_NAME";
    public static String COL_LOGGER_NAME = "LOGGER_NAME";
    public static String COL_CONTEXT = "CONTEXT";
    public static String COL_MESSAGE = "MESSAGE";
    public static String COL_EXCEPTION = "EXCEPTION";
    public static String COL_LEVEL = "LEVEL";
    public static String COL_EVENT_TIME = "EVENT_TIME";
    public static String COL_TRANSMITTED_TO_SERVER = "TRANSMITTED_TO_SERVER";
    
    public static String TABLE_PREFERENCES = "T_PREFERENCES";
    public static String COL_PREFIX_PREFERENCES= "PREFS_";
    public static String COL_VERSION = "VERSION";
    public static String COL_EAT_YOUR_OWN_DOGFOOD = "EAT_YOUR_OWN_DOGFOOD";
    public static String COL_HOST_ADDRESS = "HOST_ADDRESS";
    public static String COL_PORT_LOGMONITOR = "PORT_LOGMONITOR";
    public static String COL_PORT_TRANSMISSION = "PORT_TRANSMISSION";
    public static String COL_TRANSMIT_ALWAYS = "TRANSMIT_ALWAYS";
    public static String COL_PORT_WEBSOCKET = "PORT_WEBSOCKET";
    public static String COL_LOG_TO_DB = "LOG_TO_DB";
    public static String COL_LOG_TO_LOGMONITOR = "LOG_TO_LOGMONITOR";
    public static String COL_LOG_TO_WEBSOCKET = "LOG_TO_WEBSOCKET";
    public static String COL_LOG_TASKS_BUFFER_SIZE = "LOG_TASKS_BUFFER_SIZE";
    public static String COL_DEBUG_MODE = "DEBUG_MODE";
    public static String COL_DEFAULT_LOG_LEVEL = "DEFAULT_LOG_LEVEL";
    public static String COL_ERROR_EMAIL_TO = "ERROR_EMAIL_TO";
    
    // prepared Statements - useful if you execute them more then once
    private static Statement INSERT_INTO_LOG_EVENT_STATEMENT;
    private static Statement INSERT_INTO_PREFERENCES_STATEMENT;
    
    private static Statement UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT;
    
    private static Statement SELECT_LOG_EVENTS_STATEMENT;
    private static Statement SELECT_PREFERENCES_STATEMENT;
    private static Statement SELECT_RECORD_TO_TRANSMIT_STATEMENT;
    
    private static Statement DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT;
    
    private static Statement COUNT_ALL_LOG_EVENTS_STATEMENT;
    private static Statement COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT;
    
    // used to bind NULL values
    private static String nullValue = null;
    // to transmit NULL values as string to the server
    // server then can insert NULL
    private static String emptyString = "#!_!#";
    
    /**
     * inserts one LogEvent into the DB
     * 
     * @param bean
     * @throws DatabaseOperationFailedException
     */
    public static void insertLogEventIntoDatabase(LogEventTask bean) throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (insertLogEventIntoDatabase)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	// only ONE has r/w access to DB
    	synchronized (DATABASE_SYNC_OBJECT) {
			if (INSERT_INTO_LOG_EVENT_STATEMENT == null) {
				prepareInsertIntoLogEventStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					INSERT_INTO_LOG_EVENT_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try to create a new
					prepareInsertIntoLogEventStatement();
				}
			}
			try {
				int col = 1;
				INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getDeviceIdAsHex());
				col++;
				INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getLogstoreId());
				col++;
				if (bean.getAppName4DB() == null) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getAppName4DB());
				}
				col++;
				if (bean.getLoggerName() == null) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getLoggerName());
				}
				col++;
				if (bean.getContext() == null) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getContext());
				}
				col++;
				if (bean.getMessage() == null) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getMessage());
				}
				col++;
				if (bean.getException() == null) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getException().getMessage());
				}
				col++;
				INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.getLevel());
				col++;
				if (bean.getEventTime() == Long.MIN_VALUE) {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, nullValue);
				} else {
					INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, 
							Utility.getFormattedDate4DB(bean.getEventTime()));
				}
				col++;
				INSERT_INTO_LOG_EVENT_STATEMENT.bind(col, bean.isTransmittedToServer()?1:0);
				INSERT_INTO_LOG_EVENT_STATEMENT.execute();
				// don't close the statement - will be reused
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to INSERT_INTO_LOG_EVENT_STATEMENT "+e.getMessage(), 
							CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		} 
    }
    
    private static void prepareInsertIntoLogEventStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			INSERT_INTO_LOG_EVENT_STATEMENT = logging4bbDatabase.createStatement(
					insertIntoLogEventSql()); 
			
		} catch (DatabaseException e) {
			INSERT_INTO_LOG_EVENT_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: INSERT_INTO_LOG_EVENT_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			INSERT_INTO_LOG_EVENT_STATEMENT = null;
			notifyUser("Nullpointer CREATE: INSERT_INTO_LOG_EVENT_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		}
		// prepare the statement
		try {
			INSERT_INTO_LOG_EVENT_STATEMENT.prepare();
		} catch (DatabaseException e) {
			INSERT_INTO_LOG_EVENT_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing INSERT_INTO_LOG_EVENT_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
	}
    
    private static String insertIntoLogEventSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(INSERT_INTO);
		sb.append(Characters.SPACE); 
		sb.append(Characters.APOSTROPHE);
		sb.append(TABLE_LOG_EVENT);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE);
		sb.append(Characters.LEFT_PARENTHESIS);
    	int col = -1;
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_DEVICE_ID);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_STORE_ID);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_APP_NAME);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_LOGGER_NAME);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_CONTEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_MESSAGE);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_EXCEPTION);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_LEVEL);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_EVENT_TIME);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_TRANSMITTED_TO_SERVER);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(Characters.RIGHT_PARENTHESIS);
		sb.append(Characters.SPACE);
		
		sb.append(VALUES);
		sb.append(Characters.LEFT_PARENTHESIS);
		int i=0;
		while (i < col) {
			sb.append(Characters.QUESTION_MARK);
			sb.append(Characters.COMMA);
			i++;
		}
		sb.append(Characters.QUESTION_MARK);
		sb.append(Characters.RIGHT_PARENTHESIS);
		
    	return sb.toString();
    }
    
    /**
     * deletes transmitted records from the DB
     * 
     * @throws DatabaseOperationFailedException
     */
    public static void deleteTransmittedLogEventsFromDatabase() throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (deleteTransmittedLogEventsFromDatabase)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	// only ONE has r/w access to DB
    	synchronized (DATABASE_SYNC_OBJECT) {
			if (DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT == null) {
				prepareDeleteTransmittedLogEventsStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try to create a new
					prepareDeleteTransmittedLogEventsStatement();
				}
			}
			try {
				DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT.execute();
				// don't close the statement - will be reused
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException delete DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT "+e.getMessage(), 
							CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		} 
    }
    
    private static void prepareDeleteTransmittedLogEventsStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT = logging4bbDatabase.createStatement(
					deleteTransmittedLogEventsSql()); 
			
		} catch (DatabaseException e) {
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			notifyUser("Nullpointer CREATE: DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		}
		// prepare the statement
		try {
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT.prepare();
		} catch (DatabaseException e) {
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
	}
    
    private static String deleteTransmittedLogEventsSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(DELETE_FROM);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_LOG_EVENT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_TRANSMITTED_TO_SERVER);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(1);
    	sb.append(Characters.SPACE);
    	
    	return sb.toString();
    }
    
    // P R E F S
    /**
     * inserts the one and only preferences record into the DB
     * if no one exists
     * 
     * you can change the preferences from LogManager4BB (available at AppWorld)
     * or using another tool to access SQLite DBs from outside
     * 
     */
    public static void insertPreferencesIntoDatabase() throws DatabaseOperationFailedException {
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (insertPreferencesIntoDatabase)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	synchronized (DATABASE_SYNC_OBJECT) {
			// only ONE has r/w access to DB
			if (INSERT_INTO_PREFERENCES_STATEMENT == null) {
				prepareInsertPreferencesStatement(); 
			} else {
				// reset prepare state and clear previous bindings
				try {
					INSERT_INTO_PREFERENCES_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try a new one
					prepareInsertPreferencesStatement();
				}
			}
			try {
				int col = 1;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getVersion());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.eatYourOwnDogfood()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getHostAddress());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getPortLogmonitor());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getPortTransmission());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.transmitAlways()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getPortWebsocket());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.logToDB()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.logToLogmonitor()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.logToWebsocket()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.DEFAULT_LOG_TASKS_BUFFER_SIZE);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.isDebugMode()?1:0);
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getDefaultLogLevel());
				col++;
				INSERT_INTO_PREFERENCES_STATEMENT.bind(col, Logging4BBApp.getErrorMailTo());
				
				INSERT_INTO_PREFERENCES_STATEMENT.execute();
				// don't close the statement - will be reused
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to INSERT_INTO_PREFERENCES_STATEMENT "+e.getMessage(),
							CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		}
    }
    private static void prepareInsertPreferencesStatement() throws DatabaseOperationFailedException{
    	try {
			INSERT_INTO_PREFERENCES_STATEMENT = logging4bbDatabase.createStatement(insertPreferencesSql());
		} catch (DatabaseException e) {
			INSERT_INTO_PREFERENCES_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: INSERT_INTO_PREFERENCES_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			INSERT_INTO_PREFERENCES_STATEMENT = null;
			notifyUser("Nullpointer CREATE: INSERT_INTO_PREFERENCES_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
        try {
        	INSERT_INTO_PREFERENCES_STATEMENT.prepare();
		} catch (DatabaseException e) {
			INSERT_INTO_PREFERENCES_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing INSERT_INTO_PREFERENCES_STATEMENT "+e.getMessage(), 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }
    
    private static String insertPreferencesSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(INSERT_INTO);
		sb.append(Characters.SPACE); 
		sb.append(Characters.APOSTROPHE);
		sb.append(TABLE_PREFERENCES);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE);
		sb.append(Characters.LEFT_PARENTHESIS);
    	int col = -1;
    	
    	sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_VERSION);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_EAT_YOUR_OWN_DOGFOOD);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_HOST_ADDRESS);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_LOGMONITOR);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_TRANSMISSION);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_TRANSMIT_ALWAYS);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_WEBSOCKET);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_DB);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_LOGMONITOR);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_WEBSOCKET);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TASKS_BUFFER_SIZE);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_DEBUG_MODE);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_DEFAULT_LOG_LEVEL);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_ERROR_EMAIL_TO);
		sb.append(Characters.SPACE);
		col++;
		
		sb.append(Characters.RIGHT_PARENTHESIS);
		sb.append(Characters.SPACE);
		
		sb.append(VALUES);
		sb.append(Characters.LEFT_PARENTHESIS);
		int i=0;
		while (i < col) {
			sb.append(Characters.QUESTION_MARK);
			sb.append(Characters.COMMA);
			i++;
		}
		sb.append(Characters.QUESTION_MARK);
		sb.append(Characters.RIGHT_PARENTHESIS);
		
    	return sb.toString();
		
    }
    
    /**
     * reads the one and only preferences record
     * and sets the values (static variables) at Logging4BBApp
     * 
     * @return
     * @throws DatabaseOperationFailedException
     */
    public static boolean selectPreferencesAndUpdateProperties() throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (selectPreferencesAndUpdateProperties)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	boolean found = false;  	
        synchronized (DATABASE_SYNC_OBJECT) {
			if (SELECT_PREFERENCES_STATEMENT == null) {
				prepareSelectPreferencesStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					SELECT_PREFERENCES_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try a new one
					prepareSelectPreferencesStatement();
				}
			}
			try {
				Cursor cursor = SELECT_PREFERENCES_STATEMENT
						.getCursor();
				Row row;
				int col = 0;
				String s;
				int i;
				if (cursor.next()) {
					found = true;
					row = cursor.getRow();
					col = 0;
//					("COL_EAT_YOUR_OWN_DOGFOOD, "); // 1
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setEatYourOwnDogfood(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_HOST_ADDRESS, "); //2
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							Logging4BBApp.setHostAddress(s);
						}
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_PORT_LOGMONITOR, "); //3
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setPortLogmonitor(i);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_PORT_TRANSMISSION, "); //4
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setPortTransmission(i);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_TRANSMIT_ALWAYS, "); //5
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setTransmitAlways(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_PORT_WEBSOCKET, "); //6
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setPortWebsocket(i);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_LOG_TO_DB, "); //7
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setLogToDB(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_LOG_TO_LOGMONITOR, "); //8
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setLogToLogmonitor(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_LOG_TO_WEBSOCKET, "); //9
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setLogToWebsocket(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_LOG_TASKS_BUFFER_SIZE, "); //10
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setLogTasksBufferSize(i);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_DEBUG_MODE, "); //11
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setDebugMode(i > 0);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_DEFAULT_LOG_LEVEL, "); //12
					try {
						i = row.getInteger(col++);
						Logging4BBApp.setDefaultLogLevel(i);
					} catch (DataTypeException e) {
						// ignore
					}
//			    	("COL_ERROR_EMAIL, "); //13
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							Logging4BBApp.setErrorMailTo(s);
						}
					} catch (DataTypeException e) {
						// ignore
					}
					//
				}
				// do not close the statement - will be reused later
				// don't close the Cursor because this will close the statement, too
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to SELECT_PREFERENCES_STATEMENT "
									+ e.getMessage(), CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		}
        return found;
    }

    private static void prepareSelectPreferencesStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			SELECT_PREFERENCES_STATEMENT = logging4bbDatabase.createStatement(
					selectPreferencesSql()
					);
		} catch (DatabaseException e) {
			SELECT_PREFERENCES_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: SELECT_PREFERENCES_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			SELECT_PREFERENCES_STATEMENT = null;
			notifyUser("Nullpointer CREATE: SELECT_PREFERENCES_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
    	// prepare the statement 
    	try {
    		SELECT_PREFERENCES_STATEMENT.prepare();
		} catch (DatabaseException e) {
			SELECT_PREFERENCES_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing SELECT_PREFERENCES_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }
    private static String selectPreferencesSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(SELECT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_EAT_YOUR_OWN_DOGFOOD);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 1
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_HOST_ADDRESS);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 2
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_PORT_LOGMONITOR);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 3
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_PORT_TRANSMISSION);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 4
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_TRANSMIT_ALWAYS);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 5
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_PORT_WEBSOCKET);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 6
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_LOG_TO_DB);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 7
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_LOG_TO_LOGMONITOR);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 8
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_LOG_TO_WEBSOCKET);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 9
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_LOG_TASKS_BUFFER_SIZE);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 10
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_DEBUG_MODE);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 11
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_DEFAULT_LOG_LEVEL);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 12
    	
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_ERROR_EMAIL_TO);
    	sb.append(Characters.SPACE); // 13
    	
    	sb.append(FROM);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_PREFERENCES);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_PREFERENCES);
    	sb.append(COL_VERSION);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(Logging4BBApp.getVersion());
    	sb.append(Characters.SPACE);
    	
    	sb.append(LIMIT);
    	sb.append(Characters.SPACE);
    	sb.append(1);
    	sb.append(Characters.SPACE);

    	return sb.toString();
    }
    
    // XXXXXXXX TRANSMIT
    /**
     * gets all complete RECORDS not yet transmitted
     * 
     * @return
     * @throws DatabaseOperationFailedException 
     */
    public static LongHashtable selectRecordToTransmit() throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (selectRecordToTransmit)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	LongHashtable recordsToTransmit;  	
        synchronized (DATABASE_SYNC_OBJECT) {
			if (SELECT_RECORD_TO_TRANSMIT_STATEMENT == null) {
				prepareSelectRecordsToTransmitStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					SELECT_RECORD_TO_TRANSMIT_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try a new one
					prepareSelectRecordsToTransmitStatement();
				}
			}
			recordsToTransmit = new LongHashtable(100);		
			try {
				Cursor cursor = SELECT_RECORD_TO_TRANSMIT_STATEMENT
						.getCursor();
				Row row;
				int col = 0;
				long id;
				String s;
				StringBuffer sb;
				while (cursor.next()) {
					row = cursor.getRow();
					col = 0;
					sb = new StringBuffer();
//					("COL_ID, "); // 1
					try {
						id = row.getLong(col++);
						sb.append(id);
					} catch (DataTypeException e) {
						continue;
						// TODO log - without an ID we couldn't work
					}
					sb.append(Characters.COMMA);
//			    	("COL_DEVICE_ID, "); //2
					try {
						Object obj = row.getObject(col++);
						sb.append(obj.toString());
					} catch (DataTypeException e) {
						continue;
						// TODO log - without DEVICE ID we couldn't work
					}
					sb.append(Characters.COMMA);
//			    	("COL_STORE_ID, "); //3
					try {
						sb.append(row.getLong(col++));
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_APP_NAME, "); //4
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							sb.append(s);
						} else {
							sb.append(emptyString);
						}
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_LOGGER_NAME, "); //5
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							sb.append(s);
						} else {
							sb.append(emptyString);
						}
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_CONTEXT, "); //6
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							sb.append(s);
						} else {
							sb.append(emptyString);
						}
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_MESSAGE, "); //7
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							sb.append(s);
						} else {
							sb.append(emptyString);
						}
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_EXCEPTION, "); //8
					try {
						s = row.getString(col++);
						if (s != null && s.length() > 0) {
							sb.append(s);
						} else {
							sb.append(emptyString);
						}
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					sb.append(Characters.COMMA);
//			    	("COL_LEVEL, "); //9
					try {
						sb.append(row.getInteger(col++));
					} catch (DataTypeException e) {
						sb.append(0);
					}
					sb.append(Characters.COMMA);
//			    	("COL_EVENT_TIME, "); //10
					try {
						sb.append(row.getString(col++));
					} catch (DataTypeException e) {
						sb.append(emptyString);
					}
					//
					recordsToTransmit.put(id, sb);
				}
				// do not close the statement - will be reused later
				// don't close the Cursor because this will close the statement, too
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to SELECT_RECORD_TO_TRANSMIT_STATEMENT "
									+ e.getMessage(), CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		}
        return recordsToTransmit;
    }

    private static void prepareSelectRecordsToTransmitStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			SELECT_RECORD_TO_TRANSMIT_STATEMENT = logging4bbDatabase.createStatement(
					selectRecordsToTransmitSql()
					);
		} catch (DatabaseException e) {
			SELECT_RECORD_TO_TRANSMIT_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: SELECT_RECORD_TO_TRANSMIT_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			SELECT_RECORD_TO_TRANSMIT_STATEMENT = null;
			notifyUser("Nullpointer CREATE: SELECT_RECORD_TO_TRANSMIT_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
    	// prepare the statement 
    	try {
    		SELECT_RECORD_TO_TRANSMIT_STATEMENT.prepare();
		} catch (DatabaseException e) {
			SELECT_RECORD_TO_TRANSMIT_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing SELECT_RECORD_TO_TRANSMIT_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }

    private static String selectRecordsToTransmitSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(SELECT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_ID);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 1
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_DEVICE_ID);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 2
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_STORE_ID);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 3
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_APP_NAME);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 4
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_LOGGER_NAME);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 5
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_CONTEXT);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 6
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_MESSAGE);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 7
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_EXCEPTION);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 8
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_LEVEL);
    	sb.append(Characters.COMMA);
    	sb.append(Characters.SPACE); // 9
    	
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_EVENT_TIME);
    	sb.append(Characters.SPACE); // 10
    	
    	sb.append(FROM);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_LOG_EVENT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_TRANSMITTED_TO_SERVER);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(0);
    	sb.append(Characters.SPACE);
    	
    	sb.append(ORDER_BY);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_ID);
    	sb.append(Characters.SPACE);
    	sb.append(ASC);
    	sb.append(Characters.SPACE);
    	
    	sb.append(LIMIT);
    	sb.append(Characters.SPACE);
    	sb.append(100);
    	sb.append(Characters.SPACE);

    	return sb.toString();
    }

    /**
     * Updates Records transmitted to server
     * Starts a TX, updates all from Vector, COMMITs TX
     * 
     * @param ids Vector of IDs
     * @throws DatabaseOperationFailedException 
     */
    public static void updateLogEventsTransmitted(LongVector ids) throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (updateRecordTransmitted)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	// only ONE has r/w access to DB
    	synchronized (DATABASE_SYNC_OBJECT) {
    		try {
				logging4bbDatabase.beginTransaction();
			} catch (DatabaseException e1) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException begin TX to UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT "+e1.getMessage(),
							CRITICAL_ERROR, WORK_ON);
				}
				try {
					logging4bbDatabase.rollbackTransaction();
				} catch (DatabaseException e) {
					// TODO LOG
				}
				throw new DatabaseOperationFailedException(e1);
			}
    		long id = Long.MIN_VALUE;
    		for (int i = 0; i < ids.size(); i++) {
				id = ids.elementAt(i);
				if (id != Long.MIN_VALUE) {
					if (UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT == null) {
						prepareUpdateLogEventsTransmittedStatement();
					} else {
						// reset prepare state and clear previous bindings
						try {
							UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT.reset();
						} catch (DatabaseException e) {
							// try a new one
							prepareUpdateLogEventsTransmittedStatement();
						}
					}
					try {
						UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT.bind(1, id);
						UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT.execute();
						// don't close the statement - will be reused
					} catch (DatabaseException e) {
						if (!logging4bbDatabaseIsPaused) {
							notifyUser(
									"DatabaseException binding to UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT "+e.getMessage(),
									CRITICAL_ERROR, WORK_ON);
						}
						try {
							logging4bbDatabase.rollbackTransaction();
						} catch (DatabaseException e1) {
							// TODO LOG
						}
						throw new DatabaseOperationFailedException(e);
					}
				}
			}
    		// WORK DONE  COMMIT
    		try {
				logging4bbDatabase.commitTransaction();
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException commit TX to UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT "+e.getMessage(),
							CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
		}
    }
    private static void prepareUpdateLogEventsTransmittedStatement() throws DatabaseOperationFailedException{
    	try {
			UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT = logging4bbDatabase.createStatement(updateLogEventTransmittedSql());
		} catch (DatabaseException e) {
			UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT = null;
			notifyUser("Nullpointer CREATE: UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
        try {
        	UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT.prepare();
		} catch (DatabaseException e) {
			UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }
    
    private static String updateLogEventTransmittedSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(UPDATE);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_LOG_EVENT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(SET);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_TRANSMITTED_TO_SERVER);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(1);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_ID);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.QUESTION_MARK);
    	sb.append(Characters.SPACE);
    	
    	return sb.toString();
    }
    
    // S T A T I S T I C
    /**
     * counts all log events
     */
    public static int countAllLogEvents() throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (countAllRecords)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	int count = 0;  	
        synchronized (DATABASE_SYNC_OBJECT) {
			if (COUNT_ALL_LOG_EVENTS_STATEMENT == null) {
				prepareCountAllLogEventsStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					COUNT_ALL_LOG_EVENTS_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try a new one
					prepareCountAllLogEventsStatement();
				}
			}	
			try {
				Cursor cursor = COUNT_ALL_LOG_EVENTS_STATEMENT
						.getCursor();
				Row row;
				if (cursor.next()) {
					row = cursor.getRow();
					count = row.getInteger(0);       
				}
				// do not close the statement - will be reused later
				// don't close the Cursor because this will close the statement, too
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to COUNT_ALL_LOG_EVENTS_STATEMENT "
									+ e.getMessage(), CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
			catch (DataTypeException e) {
				// ignore
			}
		}
        return count;
    }
    private static void prepareCountAllLogEventsStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			COUNT_ALL_LOG_EVENTS_STATEMENT = logging4bbDatabase.createStatement(
					countALlLogEventsSql()
					);
		} catch (DatabaseException e) {
			COUNT_ALL_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: COUNT_ALL_LOG_EVENTS_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			COUNT_ALL_LOG_EVENTS_STATEMENT = null;
			notifyUser("Nullpointer CREATE: COUNT_ALL_LOG_EVENTS_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
    	// prepare the statement 
    	try {
    		COUNT_ALL_LOG_EVENTS_STATEMENT.prepare();
		} catch (DatabaseException e) {
			COUNT_ALL_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing COUNT_ALL_LOG_EVENTS_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }
    private static String countALlLogEventsSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(SELECT);
    	sb.append(Characters.SPACE);
    	sb.append(COUNT);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.LEFT_PARENTHESIS);
    	sb.append(Characters.ASTERISK);
    	sb.append(Characters.RIGHT_PARENTHESIS);
    	sb.append(Characters.SPACE);
    	
    	sb.append(FROM);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_LOG_EVENT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_ID);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.GREATER_THAN_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(0);
    	sb.append(Characters.SPACE);
    	return sb.toString();
    }
    
    /**
     * counts all not transmitted LogEvents
     * 
     * @return
     * @throws DatabaseOperationFailedException
     */
    public static int countNotTransmittedLogEvents() throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (countNotTransmittedRecords)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	int count = 0;  	
        synchronized (DATABASE_SYNC_OBJECT) {
			if (COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT == null) {
				prepareCountNotTransmittedLogEventsStatement();
			} else {
				// reset prepare state and clear previous bindings
				try {
					COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT.reset();
				} catch (DatabaseException e) {
					// try a new one
					prepareCountNotTransmittedLogEventsStatement();
				}
			}	
			try {
				Cursor cursor = COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT
						.getCursor();
				Row row;
				if (cursor.next()) {
					row = cursor.getRow();
					count = row.getInteger(0);       
				}
				// do not close the statement - will be reused later
				// don't close the Cursor because this will close the statement, too
			} catch (DatabaseException e) {
				if (!logging4bbDatabaseIsPaused) {
					notifyUser(
							"DatabaseException binding to COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT "
									+ e.getMessage(), CRITICAL_ERROR, WORK_ON);
				}
				throw new DatabaseOperationFailedException(e);
			}
			catch (DataTypeException e) {
				// ignore
			}
		}
        return count;
    }
    private static void prepareCountNotTransmittedLogEventsStatement() throws DatabaseOperationFailedException{
    	// create the statement
    	try {
			COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT = logging4bbDatabase.createStatement(
					countNotTransmittedLogEventsSql()
					);
		} catch (DatabaseException e) {
			COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException CREATE: COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		} catch (NullPointerException e) {
			COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			notifyUser("Nullpointer CREATE: COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT" , CRITICAL_ERROR, EXIT_THE_APP);
		} 
    	// prepare the statement 
    	try {
    		COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT.prepare();
		} catch (DatabaseException e) {
			COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
			if (!logging4bbDatabaseIsPaused) {
				notifyUser("DatabaseException preparing COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT "+e.getMessage() , 
						CRITICAL_ERROR, WORK_ON);
			}
			throw new DatabaseOperationFailedException(e);
		}
    }
    private static String countNotTransmittedLogEventsSql(){
    	StringBuffer sb = new StringBuffer();
    	sb.append(SELECT);
    	sb.append(Characters.SPACE);
    	sb.append(COUNT);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.LEFT_PARENTHESIS);
    	sb.append(Characters.ASTERISK);
    	sb.append(Characters.RIGHT_PARENTHESIS);
    	sb.append(Characters.SPACE);
    	
    	sb.append(FROM);
    	sb.append(Characters.SPACE);
    	sb.append(TABLE_LOG_EVENT);
    	sb.append(Characters.SPACE);
    	
    	sb.append(WHERE);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_ID);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.GREATER_THAN_SIGN);
    	sb.append(Characters.SPACE);
    	sb.append(0);
    	sb.append(Characters.SPACE);
    	
    	sb.append(AND);
    	sb.append(Characters.SPACE);
    	sb.append(COL_PREFIX_LOG_EVENT);
    	sb.append(COL_TRANSMITTED_TO_SERVER);
    	sb.append(Characters.SPACE);
    	sb.append(Characters.EQUALS_SIGN);
    	sb.append(0);
    	sb.append(Characters.SPACE);
    	return sb.toString();
    }
 
	//  C R E A T E    HELPERS
    /**
     * CREATES the Table for LogEvents if DB is empty/new
     * 
     */
	private static String createTableStatement4LogEventSql(){
		StringBuffer sb = new StringBuffer();
		sb.append(CREATE_TABLE);
		sb.append(Characters.SPACE); 
		sb.append(Characters.APOSTROPHE);
		sb.append(TABLE_LOG_EVENT);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE);
		sb.append(Characters.LEFT_PARENTHESIS);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_ID);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_PK);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_DEVICE_ID);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_STORE_ID);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_APP_NAME);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_LOGGER_NAME);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_CONTEXT);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_MESSAGE);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_EXCEPTION);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_LEVEL);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_EVENT_TIME);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_DATETIME);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_LOG_EVENT);
		sb.append(COL_TRANSMITTED_TO_SERVER);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.RIGHT_PARENTHESIS);
		return sb.toString();
	}
	
	private static String createTableStatement4PreferencesSql(){
		StringBuffer sb = new StringBuffer();
		sb.append(CREATE_TABLE);
		sb.append(Characters.SPACE); 
		sb.append(Characters.APOSTROPHE);
		sb.append(TABLE_PREFERENCES);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE);
		sb.append(Characters.LEFT_PARENTHESIS);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_ID);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_PK);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_VERSION);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_EAT_YOUR_OWN_DOGFOOD);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_HOST_ADDRESS);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_LOGMONITOR);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_TRANSMISSION);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_TRANSMIT_ALWAYS);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_PORT_WEBSOCKET);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_DB);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_LOGMONITOR);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TO_WEBSOCKET);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_LOG_TASKS_BUFFER_SIZE);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_DEBUG_MODE);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_DEFAULT_LOG_LEVEL);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_NUMERIC);
		sb.append(Characters.COMMA);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.APOSTROPHE);
		sb.append(COL_PREFIX_PREFERENCES);
		sb.append(COL_ERROR_EMAIL_TO);
		sb.append(Characters.APOSTROPHE);
		sb.append(Characters.SPACE); 
		sb.append(COL_TYPE_TEXT);
		sb.append(Characters.SPACE);
		
		sb.append(Characters.RIGHT_PARENTHESIS);
		return sb.toString();
	}
	
    public static void updatePreferences(String sql) throws DatabaseOperationFailedException{
    	if (!verifyDatabaseExistsAndIsOpened()) {
    		if (!logging4bbDatabaseIsPaused) {
    			notifyUser("Database doesn't exist (updatePreferences)", CRITICAL_ERROR, WORK_ON);
			}
    		throw new DatabaseOperationFailedException("");
		}
    	// only ONE has r/w access to DB
    	synchronized (DATABASE_SYNC_OBJECT) {
    		try {
				logging4bbDatabase.beginTransaction();
			} catch (DatabaseException e1) {
				try {
					logging4bbDatabase.rollbackTransaction();
				} catch (DatabaseException e) {
					// TODO LOG
				}
				throw new DatabaseOperationFailedException(e1);
			}
    		
			try {
				Statement s = logging4bbDatabase.createStatement(sql);
				s.prepare();
				s.execute();
			} catch (DatabaseException e) {
				try {
					logging4bbDatabase.rollbackTransaction();
				} catch (DatabaseException e1) {
					// TODO LOG
				}
				throw new DatabaseOperationFailedException(e);
			}
    		// WORK DONE  COMMIT
    		try {
				logging4bbDatabase.commitTransaction();
			} catch (DatabaseException e) {
				throw new DatabaseOperationFailedException(e);
			}
		}
    }

    
    /**
     * checks if all is ok with DB:
     * exists and is opened
     * if not returns false
     * 
     * @return
     */
    private static boolean verifyDatabaseExistsAndIsOpened(){
    	if (logging4bbDatabaseIsPaused) {
			return false;
		}
    	// (re)open the database if not exists
    	if (logging4bbDatabase  == null) {
			openDatabase();
			if (logging4bbDatabase == null) {
				return false;
			}
		}
    	// verify if DB can be accessed
    	try {
			logging4bbDatabase.getFile();
		} catch (DatabasePathException dpe) {
			// SD Card n/a ?
			return false;
		}
		return true;
    }
    
    public static void pauseDatabase(){
    	closeDatabase();
    	logging4bbDatabaseIsPaused = true;
    }
    
    public static void resumeDatabase(){
    	reopenDatabase();
    	logging4bbDatabaseIsPaused = false;
    }
    
    /**
     * if SD Card goes away
     * or something else happens where database has to set to null
     * 
     * we're optimistic and 
     * try to close prepared statements and
     * try to close the Database
     * 
     */
    public static void closeDatabase(){
    	synchronized (DATABASE_SYNC_OBJECT) {
			clearPreparedStatements();
			if (logging4bbDatabase != null) {
				// we're optimistic and try to close the DB
				try {
					logging4bbDatabase.close();
				} catch (DatabaseIOException e) {
				} // ignore
				logging4bbDatabase = null;
			}
		}
    }
    /**
     * if DB has to be reopened or closed (perhaps SD Card was unmounted)
     * all prepared statements should be tried to be closed and set to null,
     * because they were bound to the previous database objec
     * 
     */
    private static void clearPreparedStatements(){
    	if (INSERT_INTO_LOG_EVENT_STATEMENT != null) {
			try {
				INSERT_INTO_LOG_EVENT_STATEMENT.close();
			} catch (DatabaseException e) {}
			INSERT_INTO_LOG_EVENT_STATEMENT = null;
		}
    	if (INSERT_INTO_PREFERENCES_STATEMENT != null) {
			try {
				INSERT_INTO_PREFERENCES_STATEMENT.close();
			} catch (DatabaseException e) {}
			INSERT_INTO_PREFERENCES_STATEMENT = null;
		}
    	if (UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT != null) {
			try {
				UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT.close();
			} catch (DatabaseException e) {}
			UPDATE_LOG_EVENTS_TRANSMITTED_STATEMENT = null;
		}
    	if (SELECT_LOG_EVENTS_STATEMENT != null) {
			try {
				SELECT_LOG_EVENTS_STATEMENT.close();
			} catch (DatabaseException e) {}
			SELECT_LOG_EVENTS_STATEMENT = null;
		}
    	if (SELECT_PREFERENCES_STATEMENT != null) {
			try {
				SELECT_PREFERENCES_STATEMENT.close();
			} catch (DatabaseException e) {}
			SELECT_PREFERENCES_STATEMENT = null;
		}
    	if (SELECT_RECORD_TO_TRANSMIT_STATEMENT != null) {
			try {
				SELECT_RECORD_TO_TRANSMIT_STATEMENT.close();
			} catch (DatabaseException e) {}
			SELECT_RECORD_TO_TRANSMIT_STATEMENT = null;
		}
    	if (DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT != null) {
			try {
				DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT.close();
			} catch (DatabaseException e) {}
			DELETE_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
		}
    	if (COUNT_ALL_LOG_EVENTS_STATEMENT != null) {
			try {
				COUNT_ALL_LOG_EVENTS_STATEMENT.close();
			} catch (DatabaseException e) {}
			COUNT_ALL_LOG_EVENTS_STATEMENT = null;
		}
    	if (COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT != null) {
			try {
				COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT.close();
			} catch (DatabaseException e) {}
			COUNT_NOT_TRANSMITTED_LOG_EVENTS_STATEMENT = null;
		}
    }
    
    /**
     * try to re-open the database
     * per ex. if SD Card comes back
     * 
     */
    public static void reopenDatabase(){
    	// try to close if DB already was used before
    	closeDatabase();
    	openDatabase();
    }
	
    /**
     * opens Database
     * if wrong URI - exits the application
     * 
     */
    private static void openDatabase(){             
        URI uri = null;
		try {
			uri = URI.create(
					Logging4BBApp.LOGGING4BB_DB_LOCATION 
					+ Logging4BBApp.LOGGING4BB_DB_NAME);
		} catch (IllegalArgumentException e) {
			notifyUser("(IllegalArgumentException) Couldn't open the Logging4BB Database. Exiting application... "+e.getMessage(), 
					CRITICAL_ERROR, EXIT_THE_APP);
			return;
		} catch (MalformedURIException e) {
			notifyUser("(MalformedURIException)Couldn't open the Logging4BB Database. Exiting application... "+e.getMessage(), 
					CRITICAL_ERROR, EXIT_THE_APP);
			return;
		}     
        
        synchronized (DATABASE_SYNC_OBJECT) {
        	// double check if the database was just oppened by another thread
        	if (logging4bbDatabase != null) {
        		// job already done
				return;
			}
			// Open or create a plain text database.  This will create the
			// directory and file defined by the URI (if they do not already exist).
			try {
				logging4bbDatabase = DatabaseFactory.openOrCreate(uri,
						new DatabaseSecurityOptions(false));
			} catch (ControlledAccessException e) {
				notifyUser(
						"(ControlledAccessException)Couldn't open the Logging4BB Database. Exiting application... "
								+ e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
				return;
			} catch (DatabaseIOException e) {
				notifyUser(
						"(DatabaseIOException)Couldn't open the Logging4BB Database. Exiting application... "
								+ e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
				return;
			} catch (DatabasePathException e) {
				notifyUser(
						"(DatabasePathException)Couldn't open the Logging4BB Database. Exiting application... "
								+ e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
				return;
			} catch (IllegalArgumentException e) {
				notifyUser(
						"(IllegalArgumentException) Couldn't open the Logging4BB Database. Exiting application... "
								+ e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
				return;
			}
			createTableIfNotAvailable(TABLE_LOG_EVENT);
			createTableIfNotAvailable(TABLE_PREFERENCES);
		}
    }
    
    private static void notifyUser(String message, boolean critical, boolean exit){
    	StringBuffer sb = new StringBuffer();
    	sb.append(message);
    	if (critical) {
			sb.append("<--  is  critical");
		}
    	if (exit) {
			sb.append(" should EXIT");
		}
    	Logging4BBApp.getInstance().notifyForeground(message);
    }
    
    private static void createTableIfNotAvailable(String tableName){
    	boolean exists = false;
        String SQLTableSelect = "SELECT name FROM sqlite_master " +
                                " WHERE type='table' " +
                               " ORDER BY name";
        try {
			Statement st = logging4bbDatabase.createStatement(SQLTableSelect);
			st.prepare();
			Cursor c = st.getCursor();
			Row r;
			while ( c.next() && !exists ) {
			    r = c.getRow();
			    String name = r.getString(0);
			    if (name.equalsIgnoreCase(tableName)) {
					exists = true;
				}
			}
			c.close();
			if (!exists) {
				if (tableName.equals(TABLE_LOG_EVENT)) {
					st = logging4bbDatabase.createStatement(createTableStatement4LogEventSql());
				}  else {
					st = logging4bbDatabase.createStatement(createTableStatement4PreferencesSql());
				}
				st.prepare();
				st.execute();
			}
		} catch (DatabaseException e) {
			notifyUser("Create Table DB Exception: "+e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
		} catch (DataTypeException e) {
			notifyUser("DataTypeException: "+e.getMessage(), CRITICAL_ERROR, EXIT_THE_APP);
		}
    }
    
    /**
     * if preferences record already exists in DB: 
     * then replace parameters in Logger4BBApp with stored values
     * otherwise create record with defaults from logger4BBApp
     * I'm using only one preferences record
     * but you can rewrite this and use a preferences record for each store id
     * in my cases one global preferences record fits into my needs
     */
    public static void initPreferences(){
    	try {
			if (selectPreferencesAndUpdateProperties()) {
				logger.info("DB Preferences selected");
				return;
			} else {
				insertPreferencesIntoDatabase();
				Logging4BBApp.setLogTasksBufferSize(Logging4BBApp.DEFAULT_LOG_TASKS_BUFFER_SIZE);
				logger.info("DB Preferences inserted");
			}
		} catch (DatabaseOperationFailedException e) {
			logger.warn("couldn't init DB Preferences", e);
		}
    }
    
    /**
     * refreshes preferences from DB and restarts Logger
     */
    public static void refreshPreferences(Hashtable preferences){
    	try {
    		// update existing values in DB
        	StringBuffer sb = new StringBuffer();
        	sb.append(UPDATE);
        	sb.append(Characters.SPACE);
        	sb.append(TABLE_PREFERENCES);
        	sb.append(Characters.SPACE);
        	sb.append(SET);
        	sb.append(Characters.SPACE);
    		Object o;
    		boolean found = false;
    		o = preferences.get(COL_LOG_TO_DB);
			if (o instanceof Integer) {
				found = true;
				sb.append(COL_PREFIX_PREFERENCES);
	        	sb.append(COL_LOG_TO_DB);
	        	sb.append(Characters.SPACE);
	        	sb.append(Characters.EQUALS_SIGN);
	        	sb.append(Characters.SPACE);
	        	sb.append(((Integer)o).intValue());
	        	sb.append(Characters.SPACE);
			}
			o = preferences.get(COL_LOG_TO_LOGMONITOR);
			if (o instanceof Integer) {
				if (found) {
					sb.append(Characters.COMMA);
					sb.append(Characters.SPACE);
				} else {
					found = true;
				}
				sb.append(COL_PREFIX_PREFERENCES);
	        	sb.append(COL_LOG_TO_LOGMONITOR);
	        	sb.append(Characters.SPACE);
	        	sb.append(Characters.EQUALS_SIGN);
	        	sb.append(Characters.SPACE);
	        	sb.append(((Integer)o).intValue());
	        	sb.append(Characters.SPACE);
			}
			o = preferences.get(COL_HOST_ADDRESS);
			if (o instanceof String) {
				if (found) {
					sb.append(Characters.COMMA);
					sb.append(Characters.SPACE);
				} else {
					found = true;
				}
				sb.append(COL_PREFIX_PREFERENCES);
	        	sb.append(COL_HOST_ADDRESS);
	        	sb.append(Characters.SPACE);
	        	sb.append(Characters.EQUALS_SIGN);
	        	sb.append(Characters.SPACE);
	        	sb.append(((String)o));
	        	sb.append(Characters.SPACE);
			}	
			sb.append(WHERE);
        	sb.append(Characters.SPACE);
        	sb.append(COL_PREFIX_PREFERENCES);
        	sb.append(COL_ID);
        	sb.append(Characters.SPACE);
        	sb.append(Characters.EQUALS_SIGN);
        	sb.append(Characters.SPACE);
        	sb.append(1);
        	sb.append(Characters.SPACE);
        	if (found) {
				updatePreferences(sb.toString());
			}
        	// now select and update
			if (selectPreferencesAndUpdateProperties()) {
				CommandFactory.resumeLoggerCommand().execute(new Hashtable());
				logger.debug(
						"refresh Preferences","DONE Log2DB: "
						+Logging4BBApp.logToDB()+" Log2Socket: "
						+Logging4BBApp.logToLogmonitor());
				return;
			}
		} catch (DatabaseOperationFailedException e) {
			logger.warn("couldn't init DB Preferences", e);
		}
    }
	
}
