/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.socket;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.Vector;

import javax.microedition.io.StreamConnection;

import net.rim.device.api.compress.GZIPOutputStream;
import net.rim.device.api.io.transport.ConnectionDescriptor;
import net.rim.device.api.io.transport.ConnectionFactory;
import net.rim.device.api.io.transport.TransportInfo;
import net.rim.device.api.system.Characters;

import org.appbus.bb.app.logging4bb.Logging4BBApp;
import org.appbus.bb.app.logging4bb.api.ILogger;
import org.appbus.bb.app.logging4bb.db.Logging2DatabaseWorker;
import org.appbus.bb.app.logging4bb.logger.Logger;
import org.appbus.bb.app.logging4bb.logger.LoggerFactory;
import org.appbus.bb.app.logging4bb.model.LogEventTask;

public class Logging2LogMonitorSocketWorker {
	
	private static Vector tasks = new Vector(Logging4BBApp.DEFAULT_LOG_TASKS_BUFFER_SIZE);
	private static boolean isWorking = false;
	private static boolean isPaused = false;
	private static boolean killMe = false;
	
	private static int connectionFailedCounter = 0;
	// trying 20 times to connect
	// each rime pauses 3000 ms == 60 sec
	private static final int CONNECTION_RETRY_MAX = 20; 
	
	private static boolean haveMoreWorkToDo = false;
	private static WorkerThread theThread;
	private static boolean bufferTooSmallLogged = false;
	
	private static String UTF_8 = "utf-8";
	private static String LEVEL = "level";
	private static String TIMESTAMP_MILLIS = "timeStampMillis";
	private static String MESSAGE = "Message";
	private static String NAME = "name";
	private static String SLASH = "/";
	
    // we're using PORT_10020 TODO JSON on port 10031 for WebSockets
	private static String LOGGING_EVENT_HEADER =
        "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
	    + "<LoggingEvent "
	    + "xmlns=\"http://lilith.sf.net/schema/logging/13\" "
	    + "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" "
	    + "xsi:schemaLocation=\""
	    + "http://lilith.sf.net/schema/logging/13 "
	    + "http://lilith.sf.net/schema/logging/13/LoggingEvent.xsd\" "
	    + "logger=\""
		;
	
	private static String LOGGING_EVENT_ARGS =
		"<Arguments>"
	    + "<Argument>arg1</Argument>"
	    + "<Argument>arg2</Argument>"
	    + "<Argument>arg3</Argument>"
	    + "</Arguments>"
	    + "<LoggerContext"
		;
	
	private static String LOGGING_EVENT_PROPS =
	    ">"
	    + "<Properties>"
	    + "<Entry key = \"applicationIdentifier\">"
		;
	
	private static String LOGGING_EVENT_FOOTER =
		"</Entry>"
	    + "</Properties>"
	    + "</LoggerContext>"
	    + "</LoggingEvent>"
		;
	
	/**
	 * adds one task to the Vector
	 * synchronized on the Vector, so data is always consistent
	 * 
	 * @param task
	 */
	public static void addTask(Object task){
		if (isPaused) {
			return;
		}
		synchronized (tasks) {
			if (tasks.size() < Logging4BBApp.getLogTasksBufferSize()) {
				tasks.addElement(task);
				bufferTooSmallLogged = false;
			} else {
				// ignore task
				// max reached log error (only one time)
				if (!bufferTooSmallLogged) {
					bufferTooSmallLogged = true;
					tasks.addElement(getErrorLogEventBufferTooSmall());
				}
			}
		}
	}
	
	/**
	 * there are some special cases where transport types aren't ready
	 * after powerUp()
	 * we don't want to loose an event, so we put it back
	 * 
	 * @param task
	 */
	public static void restoreTask(Object task){
		synchronized (tasks) {
			tasks.insertElementAt(task, 0);
		}
	}
	
	/**
	 * adds a bunch of tasks to the Vector
	 * synchronized on the Vector, so data is always consistent
	 * 
	 * @param tasks
	 */
	public static void addTasks(Hashtable[] someTasks){
		if (isPaused) {
			return;
		}
		synchronized (tasks) {
			if ((tasks.size()+someTasks.length) <= Logging4BBApp.getLogTasksBufferSize()) {
				for (int i = 0; i < someTasks.length; i++) {
					tasks.addElement(new LogEventTask(someTasks[i]));
				}
				bufferTooSmallLogged = false;
			} else {
				int max = Logging4BBApp.getLogTasksBufferSize()-tasks.size();
				for (int i = 0; i <= max; i++) {
					tasks.addElement(new LogEventTask(someTasks[i]));
				}
				// ignore other tasks
				// max reached log error (only one time)
				if (!bufferTooSmallLogged) {
					bufferTooSmallLogged = true;
					tasks.addElement(getErrorLogEventBufferTooSmall());
				}
			}
		}
	}
	
	private static LogEventTask getErrorLogEventBufferTooSmall(){
		LogEventTask errorLogEvent = new LogEventTask(
				LoggerFactory.LOG_STORE_ID, 
				LoggerFactory.APP_NAME, 
				Logging2LogMonitorSocketWorker.class.getName());
		errorLogEvent.setContext("unexpected behaviour");
		errorLogEvent.setMessage("LogBuffer too small: "+Logging4BBApp.getLogTasksBufferSize());
		errorLogEvent.setLevel(ILogger.LOG_LEVEL_ERROR);
		return errorLogEvent;
	}
	
	/**
	 * tries to get the next task from the vector
	 * if Vector is empty, NoSuchElementException was thrown
	 * so the caller knows to go into sleep mode
	 * 
	 * @return
	 * @throws NoSuchElementException
	 */
	public static Object getNextTask() throws NoSuchElementException{
		if (tasks.isEmpty()) {
			throw new NoSuchElementException("Worker queue is empty");
		}
		synchronized (tasks) {
			Object task = (Object) tasks.firstElement();
			tasks.removeElementAt(0);
			return task;
		}
	}
	
	public static void startup(){
		isWorking = true;
		isPaused = false;
		theThread = new WorkerThread();
		theThread.start();
	}
	
	public static void ensureCapacity(int minimumCapacity){
		// tasks are created with minimum 500
		// this can be overwritten by startup parameters
		// or from LogManager
		synchronized (tasks) {
			tasks.ensureCapacity(minimumCapacity); 
		}
	}
	
	public static void shutdown(){
		if (theThread != null) {
			if (theThread.isAlive()) {
				theThread.pingBytes = null;
				isWorking = false;
				isPaused = false;
				theThread.interrupt();
			}
			//theThread = null;
		}
	}

	/**
	 * pauses the Thread means:
	 * sleeps 3000 ms, then tests again if still paused and should not work
	 */
	public static void pauseWorker() {
		if (theThread != null && theThread.isAlive()) {
			isPaused = true;
			isWorking = false;
		}
	}
	
	
	/**
	 * restarts the worker:
	 * if Thread is alive: stop pause, start working
	 * otherwise: start a new Thread
	 * 
	 */
	public static void restartWorker(){
		if (!Logging4BBApp.logToLogmonitor()) {
			return;
		}
		if (theThread != null && theThread.isAlive()) {
			theThread.pingBytes = null;
			isWorking = true;
			isPaused = false;
		} else {
			startup();
		}
	}

	/**
	 * Thread tests if there are Tasks to do
	 * if no tasks are there sleep 1 sec and try again
	 * 
	 * if isWorking is set to false and no more work to be done
	 * the Thread terminates
	 * 
	 * @author ekke
	 *
	 */
	public static class WorkerThread extends Thread {
	    // processing log events
	    private ByteArrayOutputStream byteArrayOutStream;
		private OutputStreamWriter outputStreamWriter;
		private StringBuffer sb;
		private GZIPOutputStream gzipStream;
		
		// ping to keep the Lilith Server alive
		protected byte[] pingBytes;
		private int transportWay = -1;
		private ConnectionDescriptor connectionDescriptor;
		private StreamConnection streamConnection;
        private OutputStream outputStream;
        private DataOutputStream dataOutputStream;
		
	   private void createPingByteArray(){
	    	LogEventTask pingEvent = new LogEventTask(
	    			Logging4BBApp.MY_OWN_STORE_ID, 
	    			Logging4BBApp.MY_OWN_APP_NAME, 
	    			"ping");
	    	pingEvent.setMessage("");
	    	pingEvent.setLevel(ILogger.LOG_LEVEL_DEBUG);
	    	pingBytes = createBytes(pingEvent, true);
	    	pingBytes = compress(pingBytes);
	    }

		public void run() {
			int ping = 0;
			try {
				while (isPaused || isWorking || haveMoreWorkToDo) {
					if (isPaused) {
						sleep(3000);
						if (isWorking) {
							isPaused = false;
						}
					} else {
						try {
							Object task = getNextTask();
							haveMoreWorkToDo = true;
							ping = 0;
							if (task instanceof LogEventTask) {
								// first time ? then pingBytes are null
								if (pingBytes == null) {
									if (!createConnectionAndStartToLog()) {
										connectionFailedCounter++;
										// reset some data
										closeStreams(streamConnection, dataOutputStream, outputStream);
										pingBytes = null;
										// restore the event
										restoreTask(task);
										if (connectionFailedCounter > CONNECTION_RETRY_MAX) {
											connectionFailedCounter = 0;
											Logging4BBApp.getInstance().notifyForeground("something wrong opening connection to Lilith ?");
											return;
										} else {
											// thread will be paused, we'll try up to CONNECTION_RETRY_MAX times
											// each time pausing for 3000 ms
											isPaused = true;
										}
									} else {
										connectionFailedCounter = 0;
									}
								}
								if (!isPaused) {
									byte[] b = createBytes((LogEventTask)task, false);
									b = compress(b);
									dataOutputStream.writeInt(b.length);
									dataOutputStream.write(b);
						            dataOutputStream.flush();
								}
							} else {
								logErrorsDirect2DB("Socket2LogMonitor: wrong Task Instance ", ILogger.LOG_LEVEL_WARN);
							}
						} catch (NoSuchElementException e) {
							if (isWorking) {
								if (pingBytes != null) {
									// to keep the connection alive we have to ping
									// curious: desktop apps don't need this to do
									ping++;
									if (ping > 60) {
										ping = 0;
										dataOutputStream.writeInt(pingBytes.length);
					    				dataOutputStream.write(pingBytes);
							            dataOutputStream.flush();
									}
								}
								// nothing to do - we sleep and try again
								sleep(1000);
							} else {
								haveMoreWorkToDo = false;
							}
						}
					}
				}
				
			} catch (IOException e) {
				logErrorsDirect2DB("Socket2LogMonitor: IO Exception "+e.getMessage(), ILogger.LOG_LEVEL_ERROR);
			} catch (InterruptedException e) {
				logErrorsDirect2DB("Socket2LogMonitor: Interrupted Exception", ILogger.LOG_LEVEL_ERROR);
			} finally {
				isWorking = false;
				isPaused = false;
				killMe = false;
				logErrorsDirect2DB("Socket2LogMonitor: finally closing connection ", ILogger.LOG_LEVEL_DEBUG);
	        	closeStreams(streamConnection, dataOutputStream, outputStream);
	            transportWay = -1;
	        } 
		}
		
		/**
		 * the DB Logger was started first
		 * so - if we have some problems while logging to socket -
		 * we can log directly to the DB
		 * if eat-jour-own-dogfood was set
		 * 
		 * @param message
		 * @param level
		 */
		private void logErrorsDirect2DB(final String message, int level){
			if (Logging4BBApp.eatYourOwnDogfood() && Logging4BBApp.logToDB()) {
				LogEventTask logEvent = new LogEventTask(
						LoggerFactory.LOG_STORE_ID, 
						LoggerFactory.APP_NAME, 
						Logging2LogMonitorSocketWorker.class.getName());
				logEvent.setContext("Socket");
				logEvent.setMessage(message);
				logEvent.setLevel(level);
				logEvent.setEventTime(System.currentTimeMillis());
				Logging2DatabaseWorker.addTask(logEvent);
			} else if (level == ILogger.LOG_LEVEL_ERROR) {
				System.err.println(message);
			} else {
				System.out.println(message);
			}
		}
		
		private boolean createConnectionAndStartToLog(){
			createPingByteArray();
			String url = Logging4BBApp.getSocketServerUrl4LogMonitor();
			System.out.println("Socket2LogMonitor: want to open Socket URL "+url);
	    	ConnectionFactory cf = new ConnectionFactory();
	    	// TODO perhaps modify to your needs
	        cf.setPreferredTransportTypes(new int[]{
					TransportInfo.TRANSPORT_MDS,
					TransportInfo.TRANSPORT_BIS_B,
					//TransportInfo.TRANSPORT_TCP_WIFI,
					//TransportInfo.TRANSPORT_TCP_CELLULAR,
					//TransportInfo.TRANSPORT_WAP2
					}); 
	        connectionDescriptor = cf.getConnection(url);
	        // the logger app is a system module
	        // at startup we wait until SD card is mounted, so logging to DB works
	        // if you want to log using SocketWorker it may be that no 
	        // TransportTypes are available, if you want to start logging very early
	        // esp. if eat-your-own-dogfood is set to true
	        if (connectionDescriptor == null) {
	        	logErrorsDirect2DB("Socket2LogMonitor: ConnectionDescriptor null ", ILogger.LOG_LEVEL_WARN);
	        	transportWay = -1;
				return false;
			}
	        streamConnection = (StreamConnection) connectionDescriptor.getConnection();         
    		transportWay = connectionDescriptor.getTransportDescriptor().getTransportType();
            try {
				outputStream = streamConnection.openOutputStream();
			} catch (IOException e) {
				logErrorsDirect2DB("Socket2LogMonitor: IO Exception Open OutputStream "+e.getMessage(), ILogger.LOG_LEVEL_ERROR);
				return false;
			} 
            dataOutputStream = new DataOutputStream(outputStream);
            logErrorsDirect2DB("Socket2LogMonitor: Connection opened, Streams created", ILogger.LOG_LEVEL_DEBUG);
            return true;
		}
		
	    private void closeStreams(StreamConnection streamConnection, 
    			DataOutputStream dataOutputStream, OutputStream outputStream){
	    	if (streamConnection != null) {
	    		try {
					streamConnection.close();
				} catch (IOException e) {
					logErrorsDirect2DB(
							"SocketLoggerThread (StreamConnection) "
							+Logging4BBApp.getSocketServerUrl4LogMonitor()
							+ e.getMessage(), ILogger.LOG_LEVEL_ERROR);
				}
			}
	    	if (dataOutputStream != null) {
	    		try {
	    			dataOutputStream.close();
				} catch (IOException e) {
					logErrorsDirect2DB(
							"SocketLoggerThread (DataOutputStream) "
							+Logging4BBApp.getSocketServerUrl4LogMonitor()
							+ e.getMessage(), ILogger.LOG_LEVEL_ERROR);
				}
			}
	    	if (outputStream != null) {
	    		try {
	    			outputStream.close();
				} catch (IOException e) {
					logErrorsDirect2DB(
							"SocketLoggerThread (OutputStream) "
							+Logging4BBApp.getSocketServerUrl4LogMonitor()
							+ e.getMessage(), ILogger.LOG_LEVEL_ERROR);
				}
			}
	    }
    
	    /**
	     * transforms the LogEvent into a ByteArrayOutputStream
	     * encoded as UTF-8
	     * 
	     * @param loggerEvent
	     * @param isPing
	     * @return
	     */
	    private byte[] createBytes(LogEventTask loggerEvent, boolean isPing){
	    	byteArrayOutStream = new ByteArrayOutputStream();
	    	byte[] bytes=null;
	    	try {
	    		outputStreamWriter = new OutputStreamWriter(byteArrayOutStream, UTF_8);
	    		sb = new StringBuffer();
	    		sb.append(LOGGING_EVENT_HEADER);
	            sb.append(loggerEvent.getLoggerName());
	            sb.append(Characters.QUOTATION_MARK);
	            sb.append(Characters.SPACE);
	            sb.append(LEVEL);
	            sb.append(Characters.EQUALS_SIGN);
	            sb.append(Characters.QUOTATION_MARK);
	            sb.append(Logger.logLevelTexts[loggerEvent.getLevel()]);
	            sb.append(Characters.QUOTATION_MARK);
	            sb.append(Characters.SPACE);
	            // if ping event we extract the timestamp
	            // now Lilith receives the message
	            // the connection is alive
	            // but without a timestamp the message wasn't displayed
	            if (!isPing) {
	            	sb.append(TIMESTAMP_MILLIS);
	            	sb.append(Characters.EQUALS_SIGN);
		            sb.append(Characters.QUOTATION_MARK);
	                sb.append(loggerEvent.getEventTime());
				}
	            sb.append(Characters.QUOTATION_MARK);
	            sb.append(Characters.GREATER_THAN_SIGN);
	            sb.append(Characters.LESS_THAN_SIGN);
	            sb.append(MESSAGE);
	            sb.append(Characters.GREATER_THAN_SIGN);
	            sb.append(loggerEvent.getMessage());
	            sb.append(Characters.LESS_THAN_SIGN);
	            sb.append(SLASH);
	            sb.append(MESSAGE);
	            sb.append(Characters.GREATER_THAN_SIGN);
	            sb.append(LOGGING_EVENT_ARGS);
	            if (loggerEvent.getContext() != null && loggerEvent.getContext().length() > 0) {
					sb.append(Characters.SPACE);
					sb.append(NAME);
					sb.append(Characters.EQUALS_SIGN);
		            sb.append(Characters.QUOTATION_MARK);
					sb.append(loggerEvent.getContext());
					sb.append(Characters.QUOTATION_MARK);
				}
	            sb.append(LOGGING_EVENT_PROPS);
	            sb.append(loggerEvent.getAppName4LogMonitor());
	            if (transportWay != -1) {
	            	sb.append(TransportInfo.getTransportTypeName(transportWay));
				}
	            sb.append(LOGGING_EVENT_FOOTER);
	            String data = sb.toString();
	            outputStreamWriter.write(data, 0, data.length());
	    		outputStreamWriter.flush(); // this is
	    		outputStreamWriter.close(); // absolutely necessary!!
	    		bytes=byteArrayOutStream.toByteArray();
	    		
	    	}
	    	catch(UnsupportedEncodingException e){
	    		logErrorsDirect2DB("UnsupportedEncodingException "+ e.getMessage(), ILogger.LOG_LEVEL_ERROR);
	    	}
	    	catch(IOException e) {
	    		logErrorsDirect2DB("IOException processing event "+ e.getMessage(), ILogger.LOG_LEVEL_ERROR);
	    	}
	    	finally {
	    		if(outputStreamWriter != null){
	    			try {
	    				outputStreamWriter.close();
	    			} catch(IOException ex){
	    				// ignore
	    			}
	    		}
	    	}
	    	return bytes;
	    }
    
	    private byte[] compress( byte[] data ){    
	        try {
	            byteArrayOutStream = new ByteArrayOutputStream();
	            gzipStream = new GZIPOutputStream( 
	            		byteArrayOutStream, 6, GZIPOutputStream.MAX_LOG2_WINDOW_LENGTH );
	            gzipStream.write( data );
	            gzipStream.flush();
	            gzipStream.close();
	        }
	        catch(IOException ioe) {
	            return data;
	        }
	        return byteArrayOutStream.toByteArray();
	    }
	}
	
}
