/**
 *  Copyright (c) 2008...2012 Ekkehard Gentz (Rosenheim, Germany) and others.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *  
 *  Contributors:
 *          Ekkehard Gentz     - initial contribution 
 *  based on a snippet found at: 
 *  http://www.snippetit.com/2009/08/java-format-long-integer-into-hexadecimal-string/
 *     
 *   more info: see folder [project]/copyright
 */
package org.appbus.bb.app.logging4bb.util;

/**
 * converts long values into Strings of fixed length 16 as HEX value
 * and also back fromString into long
 * 
 * conversion methods throw NumberFormatException
 * 
 * The Java API itself does not have a class or function to convert a long into a 
 * fixed 16 characters hexadecimal string or an integer into  fixed 8 characters hexadecimal string.
 * The Java's Long.toHexString(long i) or Integer.toHexString(int i) 
 * only able to convert the integer value to a string of ASCII digits in hexadecimal (base 16) 
 * with no extra leading 0s and it does not support the reverse way (from hexadecimal string to long integer value).
 * 
 * The HexUtil class can be easily modified to support integer and short value 
 * by changing the data type and its size 
 * (e.g. integer has 4 bytes or 8 characters of hexadecimal 
 * and short has 2 bytes or 4 characters of hexadecimal)
 * 
 * To convert long integer into 16 characters hexadecimal string:
 * String result = Hex.fromLong(1234567890);
 * 
 * To convert  16 characters hexadecimal string into long integer:
 * long result = Hex.toLong("00000000499602D2");
 * 
 * @author ekke
 * 
 * EPL Eclipse Public License
 *
 */
public class HexUtil {
	private final static char[] HEX = {
	    '0', '1', '2', '3', '4', '5', '6', '7',
	    '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
	  };

	/**
	 * converts the String into a long
	 * string must be of length 16 and is only allowed to have HEX values
	 * 
	 * @param hexadecimal
	 * @return
	 * @throws NumberFormatException
	 */
	  public static long toLong(String hexadecimal) throws NumberFormatException{
	    char[] chars;
	    char c;
	    long value;
	    int i;
	    byte b;

	    if (hexadecimal == null)
	      throw new IllegalArgumentException();

	    chars = hexadecimal.toUpperCase().toCharArray();
	    if (chars.length != 16)
	      throw new NumberFormatException("Incomplete hex value");

	    value = 0;
	    b = 0;
	    for (i = 0; i < 16; i++) {
	      c = chars[i];
	      if (c >= '0' && c <= '9') {
	        value = ((value << 4) | (0xff & (c - '0')));
	      } else if (c >= 'A' && c <= 'F') {
	        value = ((value << 4) | (0xff & (c - 'A' + 10)));
	      } else {
	        throw new NumberFormatException("Invalid hex character: " + c);
	      }
	    }
	    return value;
	  }

	  /**
	   * converts a long into a String of fixed length of 16
	   * containing the HEX value filled with leading zero
	   * 
	   * @param value
	   * @return
	   */
	  public static String fromLong(long value) {
	    char[] hexs;
	    int i;
	    int c;

	    hexs = new char[16];
	    for (i = 0; i < 16; i++) {
	      c = (int)(value & 0xf);
	      hexs[16-i-1] = HEX[c];
	      value = value >> 4;
	    }
	    return new String(hexs);
	  }	
}
